# vim: expandtab tabstop=4 shiftwidth=4

from numpy import array as nparray
from numpy import floor, ceil, sqrt, log2, log10, arange
from matplotlib import pyplot as plt

from .decimal_convert_to_base import decimal_convert_to_base
from .vector_power import vector_power

def constellation_plot(constellation: nparray, binary_mode: int=0, scale: int=100):
    """
    Plots symbol numbers at corresponding constellation points

    Parameters
    ----------
    constellation:
       QAM constellation as a row numpy array vector of complex numbers
    binary_mode:
       Mode for printing symbols: 0 = decimals (default), 1 = binary
    scale:
       Size of characters on plot (default 100)

    Returns power normalized constellation as a numpy array
    """

    # Ensure constellation is a 1-dimensional complex array
    con = constellation.flatten().astype(complex)
    # Constellation size
    con_size = len(con)

    # Check Parameters
    # Constellations must contain at least 1 element
    if con_size < 1:
        print("constellation_plot error: Constellation size (",con_size,") must be at least 1.")
        return constellation

    # If user provided mode, then it must be 0 or 1
#    if (binary_mode != 0) and (binary_mode != 1):
    if binary_mode not in (0,1):
        print("constellation_plot error: User input mode (",binary_mode,") must be either 0 or 1.")
        return constellation

    # If user provided sceale, then it must be nonnegative
    if scale < 0:
        print("constellation_plot error: User input scale (",scale,") must be nonnegative.")
        return constellation

    # Number of bits per symbol
    bps = ceil(log2(con_size)).astype(int)

    # Plotting logic
    plt.suptitle("Scatterplot of {0}-point constellation".format(con_size))
    for k in arange(0,con_size):
        # Determine plot scale
        if binary_mode == 0: # Decimal Mode
            # Number of decimal digits in k
            if k == 0:
                ndigit = 1
            else:
                ndigit = floor(log10(k))+1
            size = scale*ndigit
            txt = "${0}$".format(k)
        elif binary_mode == 1: # Binary Mode
            size = scale*bps
            tmp = decimal_convert_to_base(k,2,bps).flatten()
            # Convert binary array to string
            txt = "$"
            for idx in arange(0,bps):
                txt = "{0}{1}".format(txt,tmp[idx])
            txt = "{0}$".format(txt)
        plt.scatter(con[k].real,con[k].imag,s=size,marker=txt,c='k')
    plt.show()

    # Return power normalized constellation
    return (con/sqrt(vector_power(con))).flatten()

def conplot(*args, **kwargs) -> nparray:
    '''
    Alias for constellation_plot.
    '''
    return constellation_plot(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from .complex_sinusoid import complex_sinusoid
from .constellation_plot import conplot, constellation_plot
from .data_types import FullDataType
from .exceptions import DataTypeException, EndiannessException, NumberSpaceException, SignalTypeException, WriteModeException
from .exceptions import DSPFTWException
from .filename_load import filename_load, fnload
from .generate_qam import generate_qam
from .load_bits import load_bits, loadbits
from .load_data import load_data, loaddata
from .load_signal import load_signal, loadsig
from .plot_complex import plotc, plot_complex
from .plot_3d_complex import plot_3d_complex, plot3c
from .save_bits import save_bits, savebits
from .save_data import save_data, savedata
from .save_signal import save_signal, savesig
from .signal_correlation import sigcorr, signal_correlation
from .signal_types import FullSignalType
from .truncation import true_one, true_zero
from .root_raised_cosine_filter_generator import root_raised_cosine_filter_generator, rrcfiltgen
from .root_raised_cosine import root_raised_cosine, rrcfunct
from .generate_random_qpsk import generate_random_qpsk
from .test_qpsk_modulation import test_qpsk_modulation



# vim: expandtab tabstop=4 shiftwidth=4

from numpy import ndarray

import matplotlib.pyplot as plt

def plot_complex(*args, **kwargs):
    '''
    Plots complex data in the complex plane.

    Parameters
    ----------
    args: array_like
        The complex arrays to plot
    kwargs: dict
        Parameters passed through to plt.Figure.scatter().

    '''
    plotargs = []

    for arg in args:
        if type(arg) is ndarray:
            plotargs.append(arg.real)
            plotargs.append(arg.imag)
        else:
            plotargs.append(arg)

    plt.plot(*plotargs, **kwargs)


def plotc(*args, **kwargs):
    '''
    An alias of plot_complex().
    '''
    return plot_complex(*args, **kwargs)



import numpy as np
from numpy import array as nparray

def decimal_convert_to_base(decimal_number,base,num_digits):
    """
    Given an array of decimal integers, computes an array of numbers with given base
    """

    # Ensure we have integers
    decnum = np.rint(decimal_number).astype(int)
    # Exponent array
    expArr = np.kron(np.ones(decnum.shape,dtype=int),base**np.arange(num_digits-1,-1,-1,dtype=int))
    # Base array
    baseArr = np.kron(decnum,np.ones((1,num_digits),dtype=int))

    # Return
    return ((baseArr//expArr)%base)

def num2base(*args, **kwargs) -> nparray:
    '''
    Alias for decimal_convert_to_base.
    '''
    return decimal_convert_to_base(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from .exceptions import DSPFTWException

import matplotlib.pyplot as plt
import numpy as np

def plot_3d_complex(*args, **kwargs):
    '''
    Plots complex data in a 3D space (time vs complex plane).
    
    Parameters
    ----------
    args: real and complex ndarray(s) and possibly str(s)
        The complex data along with possibly time as well as formatting string(s)
        You may pair a real array with a complex one if you would like the real
        values to be displayed on the "time" axis.  The strings are for plotting
        parameters such as "r*-" which means plot it with a red line, marking the
        points with stars and connecting the stars.
    kwargs: dict
        Parameters passed through to plt.axes.plot3D()
    '''
    # Check the input parameters in args to make sure that they're valid
    num_strs = 0
    num_reals = 0
    num_cplxs = 0
    for arg in args:
        if type(arg) is str:
            num_strs += 1
        elif np.isrealobj(arg):
            num_reals += 1
        elif np.iscomplexobj(arg):
            num_cplxs += 1
        else:
            raise DSPFTWException("Input number {} (argument {}) is not a valid input argument".format(args.index(arg), arg))
    if num_reals > num_cplxs:
        raise DSPFTWException("Too many real arguments were given.  You must have 1 complex argument for each real argument")
    if num_strs > num_cplxs:
        raise DSPFTWException("Too many formatting arguments were given.  You must have 1 complex argument for each formatting argument")

    # Check the order of the arguments to make sure they're valid
    for k in range(len(args)):
        if type(args[k]) is str:
            if k == 0:
                raise DSPFTWException("A complex argument must precede the formatting argument {}".format(args[k]))
            elif not np.iscomplexobj(args[k-1]):
                raise DSPFTWException("A complex argument must precede the formatting argument {}".format(args[k]))
        if np.isrealobj(args[k]) and type(args[k]) is not str:
            if k == len(args)-1:
                raise DSPFTWException("A complex argument must follow the real argument {}".format(args[k]))
            elif not np.iscomplexobj(args[k+1]):
                raise DSPFTWException("A complex argument must follow the real argument {}".format(args[k+1]))

    # Work through the input argument list, separating it into chunks that go together
    t_plotargs = []
    t_args = []
    k = 0
    while k < len(args):
        if np.isrealobj(args[k]):
            t_args.append(args[k])
            t_args.append(args[k+1])
            if k+2 < len(args) and type(args[k+2]) is str:
                t_args.append(args[k+2])
                k += 3
            else:
                t_args.append("-")
                k += 2
        elif np.iscomplexobj(args[k]):
            t_args.append(np.arange(args[k].shape[0]))
            t_args.append(args[k])
            if k+1 < len(args) and type(args[k+1]) is str:
                t_args.append(args[k+1])
                k += 2
            else:
                t_args.append("-")
                k += 1
        t_plotargs.append(t_args)
        t_args = []

    # Expand all the complex arrays
    plotargs = []
    for args in t_plotargs:
        if len(args[1].shape) > 1:
            for k in range(args[1].shape[1]):
                plotargs.extend([args[0], args[1][:, k].real, args[1][:, k].imag, args[2]])
        else:
            plotargs.extend([args[0], args[1].real, args[1].imag, args[2]])

    # x = time
    # y = real part
    # z = imag part
    ax = plt.axes(projection='3d')
    for k in range(len(plotargs)//4):
        t_args = plotargs[k*4:(k+1)*4]
        ax.plot(*t_args, **kwargs)
    


def plot3c(*args, **kwargs) -> plt.Figure:
    '''
    An alias of plot_complex().
    '''
    return plot_3d_complex(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), 'r') as f:
    long_description = f.read().strip()

with open(path.join(this_directory, 'VERSION'), 'r') as f:
    VERSION = f.read().strip()

setup(
    name='dspftw',
    version=VERSION,
    author='Bill Allen',
    author_email='photo.allen@gmail.com',
    description='Utilities for digital signal processing (DSP) fundamentals.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    keywords='dsp signal processing plotting complex'.split(),
    url='https://github.com/dspftw/dspftw',
    packages=['dspftw'],
    package_data={'dspftw': ['doc/*']},
    install_requires=['matplotlib', 'numpy', 'scipy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License'
    ]
)



# vim: expandtab tabstop=4 shiftwidth=4

from io import BufferedWriter
from numpy import array as nparray
from typing import Iterable

def bools_to_bytes(bools: Iterable[bool]) -> Iterable[bytes]:
    for boolean in bools:
        if boolean:
            yield b'\x01'

        else:
            yield b'\x00'

def pack_bools_to_bytes(bools: Iterable[bool]) -> Iterable[bytes]:
    offset = 0
    current_byte = 0x00

    for boolean in bools:
        current_bit = 0x00

        if boolean == True:
            current_bit = 0x01

        current_byte = current_byte | (current_bit << (7 - (offset % 8)))
        offset += 1

        if offset % 8 == 0:
            yield current_byte.to_bytes(1, 'big')
            current_byte = 0x00

    if offset % 8 != 0:
        yield current_byte.to_bytes(1, 'big')

def save_bits(file_name: str, bool_array: nparray, packed=True) -> int:
    '''
    Save bits to a file from a bool array.

    Parameters
    ----------
    file_name: string
        The name of the file to save.
    bool_array: numpy.array
        The bool array.
    packed: bool
        Whether to pack the bits into bytes.
        Defaults to True.

    Returns the number of bytes saved.
    '''
    with open(file_name, 'wb') as bit_file:
        writer = BufferedWriter(bit_file)
        count = 0

        if packed:
            for byte in pack_bools_to_bytes(bool_array):
                writer.write(byte)
                count += 1

        else:
            for byte in bools_to_bytes(bool_array):
                writer.write(byte)
                count += 1

        writer.flush()

    return count

def savebits(*args, **kwargs) -> int:
    '''
    Alias for save_bits.
    '''
    return save_bits(*args, **kwargs)
